# -*- coding: UTF-8 -*-

import os
from typing import Union, Any

import numpy as np
from pandas import DataFrame

from .. import util as ul
from ..util import path, collection, get_real_predict_label, type_20_colors, type_50_colors, plot_start, plot_end

__name__: str = "plot_pie"


def base_pie(
    values: list,
    labels: list,
    x_name: str = None,
    y_name: str = None,
    title: str = None,
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    pct_distance: float = 0.6,
    label_distance: float = 1.1,
    colors: list = None,
    autopct: str = '%1.2f%%',
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:
    fig, ax = plot_start(width, height, bottom, output, show)

    size = len(values)

    if size is not len(labels):
        ul.log(__name__).error(f"The parameter lengths of `values`({size}) and `labels`({len(labels)}) must be equal.")
        raise ValueError(f"The parameter lengths of `values`({size}) and `labels`({len(labels)}) must be equal.")

    if colors is None:
        colors = type_20_colors[:len(labels)] if size <= 20 else type_50_colors[:len(labels)]

    ax.set_axis_off()

    ax.pie(
        values,
        labels=labels,
        colors=colors,
        autopct=autopct,
        labeldistance=label_distance,
        pctdistance=pct_distance,
        **kwargs
    )

    ax.axis('off')

    plot_end(fig, title, x_name, y_name, output, show, close)


def pie_label(
    df: DataFrame,
    map_cluster: Union[str, collection],
    value: str = "value",
    clusters: str = "clusters",
    x_name: str = None,
    y_name: str = None,
    title: str = None,
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    radius: float = 0.6,
    fontsize: float = 17,
    pct_distance: float = 0.6,
    label_distance: float = 1.1,
    colors: list = None,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:
    fig, ax = plot_start(width, height, bottom, output, show)

    # judge
    df_columns = list(df.columns)

    if value not in df_columns:
        ul.log(__name__).error(
            f"The `value` ({value}) parameter must be in the `df` parameter data column name ({df_columns})")
        raise ValueError(
            f"The `value` ({value}) parameter must be in the `df` parameter data column name ({df_columns})"
        )

    df_sort, cluster_size, cluster_list = get_real_predict_label(
        df=df,
        map_cluster=map_cluster,
        clusters=clusters,
        value=value
    )

    # top value
    top_predict_cluster = list(df_sort["true_label"])[:cluster_size]
    top_x = [top_predict_cluster.count(1), top_predict_cluster.count(0)]

    if colors is None:
        colors = type_20_colors[:2]

    top_sum = np.array(top_x).sum()

    ax.set_axis_off()
    ax.pie(
        top_x,
        labels=[", ".join(cluster_list), "Other"],
        colors=colors,
        startangle=90,
        labeldistance=label_distance,
        pctdistance=pct_distance,
        wedgeprops=dict(linewidth=0),
        **kwargs
    )
    ax.pie(
        [np.array(top_x).sum()],
        colors=['white'],
        radius=radius,
        startangle=90,
        wedgeprops=dict(width=radius, edgecolor='w', linewidth=0),
        **kwargs
    )
    ax.text(0, 0, "{:.2f}%".format(top_x[0] / top_sum * 100), ha='center', va='center', fontsize=fontsize)
    ax.legend(loc='upper right')

    ax.axis('off')

    plot_end(fig, title, x_name, y_name, output, show, close)


def pie_trait(
    trait_df: DataFrame,
    trait_cluster_map: dict,
    trait_name: str = "All",
    clusters: str = "clusters",
    trait_column_name: str = "id",
    value: str = "value",
    x_name: str = None,
    y_name: str = None,
    title: str = None,
    width: float = 2,
    height: float = 2,
    radius: float = 0.6,
    fontsize: float = 17,
    pct_distance: float = 0.6,
    label_distance: float = 1.1,
    colors: list = None,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:
    trait_cluster_map_key_list = list(trait_cluster_map.keys())

    data: DataFrame = trait_df.copy()

    def trait_plot(trait_: str, atac_cell_df_: DataFrame) -> None:
        """
        show plot
        :param trait_: trait name
        :param atac_cell_df_:
        :return: None
        """
        if trait_ not in trait_cluster_map_key_list:
            ul.log(__name__).error(
                f"The key in `trait_cluster_map` does not contain the `{trait_}` trait and needs to be added")
            raise ValueError(
                f"The key in `trait_cluster_map` does not contain the `{trait_}` trait and needs to be added"
            )

        ul.log(__name__).info("Plotting pie {}".format(trait_))
        # get gene score
        trait_score = atac_cell_df_[atac_cell_df_[trait_column_name] == trait_]
        # Sort gene scores from small to large
        pie_label(
            df=trait_score[[trait_column_name, clusters, value]],
            map_cluster=trait_cluster_map[trait_],
            value=value,
            clusters=clusters,
            x_name=x_name,
            y_name=y_name,
            width=width,
            height=height,
            radius=radius,
            fontsize=fontsize,
            pct_distance=pct_distance,
            label_distance=label_distance,
            colors=colors,
            title=f"{title} {trait_}" if title is not None else title,
            output=os.path.join(output, f"cell_{trait_}_score_pie.pdf") if output is not None else None,
            show=show,
            close=close,
            **kwargs
        )

    # noinspection DuplicatedCode
    trait_list = list(set(data[trait_column_name]))
    # judge trait
    if trait_name != "All" and trait_name not in trait_list:
        ul.log(__name__).error(f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}.")
        raise ValueError(f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}.")

    # plot
    if trait_name == "All":
        for trait in trait_list:
            trait_plot(trait, trait_df)
    else:
        trait_plot(trait_name, trait_df)



# -*- coding: UTF-8 -*-

import os
from typing import Optional, Literal, List

import pandas as pd
from pandas import DataFrame

from .. import util as ul
from ..util import chrtype

__name__: str = "preprocessing_gencode"

_Feature = Optional[Literal['UTR', 'transcript', 'exon', 'Selenocysteine', 'CDS', 'start_codon', 'gene', 'stop_codon']]


def get_gene_anno(
    feature: _Feature = None,
    gtf_file: Optional[str] = None,
    filter_chromatin: bool = True,
    columns: Optional[List[str]] = None,
) -> DataFrame:

    from gtfparse import read_gtf

    cache_path: str = os.path.join(ul.project_cache_path, "gencode")
    ul.file_method(__name__).makedirs(cache_path)

    _gtf_file_: str

    if gtf_file is not None:
        if not gtf_file.endswith("gtf") and not gtf_file.endswith("gtf.gz"):
            raise ValueError("GTF files must end with .gtf or .gtf.gz.")
        else:
            _gtf_file_ = gtf_file
    else:
        filename = os.path.join(ul.project_cache_path, "gencode.annotation.gtf.gz")
        ul.log(__name__).info(f"Downloading GTF file: {filename}")

        if not os.path.exists(filename):
            ul.file_method(__name__).download_file(
                url="https://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_human/release_46/gencode.v46.annotation.gtf.gz",
                filename=os.path.join(ul.project_cache_path, "gencode.annotation.gtf.gz")
            )

        _gtf_file_ = filename
        del filename

    del gtf_file

    ul.log(__name__).info("Read GTF file")
    gtf = read_gtf(_gtf_file_)
    data = pd.DataFrame(data=gtf, columns=gtf.columns)

    if feature is not None:
        data = data[data["feature"] == feature]

    if filter_chromatin:
        chr_list: list = list(chrtype.categories)
        data = data[data["seqname"].isin(chr_list)]

    _columns_: list[str] = ['seqname', 'source', 'feature', 'start', 'end', 'strand', 'gene_id', 'gene_type', 'gene_name']

    if columns is not None:
        _columns_ = columns

    data = data[_columns_]

    return data



# -*- coding: UTF-8 -*-

from typing import Literal, Any

import numpy as np
from anndata import AnnData
import seaborn as sns
from tqdm import tqdm

from .. import util as ul
from ..tool import down_sampling_data
from ..util import path, check_adata_get, plot_end, plot_start

__name__: str = "plot_kde"


def kde(
    adata: AnnData,
    layer: str = None,
    x_name: str = None,
    y_name: str = None,
    title: str = None,
    width: float = 4,
    height: float = 2,
    bottom: float = 0.3,
    axis: Literal[-1, 0, 1] = -1,
    sample_number: int = 1000000,
    is_legend: bool = True,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
) -> None:
    ul.log(__name__).info("Start plotting the Kernel density estimation chart")

    fig, ax = plot_start(width, height, bottom, output, show)

    data = check_adata_get(adata, layer=layer, is_dense=True, is_matrix=False)

    sns.set_theme(style="whitegrid")

    # Random sampling
    if axis == -1:
        matrix = down_sampling_data(data.X, sample_number)
        sns.kdeplot(matrix, fill=True)
    elif axis == 0:
        col_number = data.shape[1]
        if data.shape[0] * data.shape[1] > sample_number:
            row_number: int = sample_number // col_number

            for i in tqdm(range(col_number)):
                _vector_ = down_sampling_data(data.X[:, i], row_number)
                sns.kdeplot(np.array(_vector_).flatten(), fill=True, **kwargs)
        else:
            for i in tqdm(range(col_number)):
                sns.kdeplot(np.array(data.X[:, i]).flatten(), fill=True, **kwargs)

        if is_legend:
            ax.legend(list(adata.var.index))

    elif axis == 1:
        row_number = data.shape[0]
        if data.shape[0] * data.shape[1] > sample_number:
            col_number: int = sample_number // row_number

            for i in tqdm(range(row_number)):
                _vector_ = down_sampling_data(data.X[i, :], col_number)
                sns.kdeplot(np.array(_vector_).flatten(), fill=True, **kwargs)
        else:
            for i in tqdm(range(row_number)):
                sns.kdeplot(np.array(data.X[i, :]).flatten(), fill=True, **kwargs)

        if is_legend:
            ax.legend(list(adata.obs.index))

    plot_end(fig, title, x_name, y_name, output, show, close)



# -*- coding: UTF-8 -*-

from typing import Any

import numpy as np
import seaborn as sns
from pandas import DataFrame

from ..util import path, plot_end, plot_start

__name__: str = "plot_bubble"


def bubble(
    df: DataFrame,
    x: str,
    y: str,
    hue: str = None,
    size: str = None,
    x_name: str = None,
    y_name: str = None,
    title: str = None,
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    output: path = None,
    show: bool = True,
    close: bool = False,
    **kwargs: Any
):

    fig, ax = plot_start(width, height, bottom, output, show)

    if size is not None:
        _size_ = df[size].values
        sizes = (np.array(_size_).min(), np.array(_size_).max())
    else:
        sizes = None

    sns.relplot(
        data=df,
        x=x,
        y=y,
        hue=hue,
        size=size,
        sizes=sizes,
        alpha=.5,
        palette="muted",
        height=6,
        **kwargs
    )

    plot_end(fig, title, x_name, y_name, output, show, close)



# -*- coding: UTF-8 -*-

import os.path
import pickle
from tqdm import tqdm
from pathlib import Path
from typing import Literal, Optional

import h5py
import pandas as pd
from pandas import DataFrame
from anndata import AnnData

from .. import util as ul
from ..util import path, to_sparse, chrtype, check_adata_get

__name__: str = "file_write"

_Field = Optional[Literal['real', 'complex', 'pattern', 'integer']]


def save_h5ad(data: AnnData, file: path) -> AnnData:
    """
    Save AnnData data
    :param data: data
    :param file: save file
    :return: AnnData file
    """
    ul.log(__name__).info("Saving data to {}".format(file))
    return data.write_h5ad(Path(file), compression='gzip')


def save_h5(data: dict, save_file: path, group_name: str = "matrix") -> None:
    """
    Save H5 data.
    :param data: data
    :param save_file: save file
    :param group_name: group name
    :return: H5 file
    """
    h5_dict = dict(data)

    file = h5py.File(f"{str(save_file)}", 'w')
    grp = file.create_group(group_name)

    for key, value in h5_dict.items():
        grp.create_dataset(key, data=value)

    file.close()


def save_pkl(data, save_file: path, is_verbose: bool = False) -> None:
    """
    Save pkl data.
    :param data: data
    :param save_file: save file
    :param is_verbose: Set true to print log;
    :return: pkl file
    """
    if is_verbose:
        ul.log(__name__).info("Saving data to {}".format(save_file))

    with open(str(save_file), 'wb') as f:
        pickle.dump(data, f) # type: ignore


def to_meta(
    adata: AnnData,
    dir_path: path,
    layer: str = None,
    feature_name: str = "peaks.bed",
    field: _Field = None
) -> None:
    """
    Convert AnnData data into metadata, i.e directory with matrix, bed file, etc.
    :param adata: input AnnData object;
    :param dir_path: Path for generating data;
    :param layer: The layer of data that needs to form meta files;
    :param feature_name: Feature file name;
    :param field: None or _Field;
        Either 'real', 'complex', 'pattern', or 'integer'.
    :return: Directory with matrix, bed file, etc.
    """

    dir_path = str(dir_path)
    ul.file_method(__name__).makedirs(dir_path)

    # Convert dense matrices to sparse matrices
    sparse_matrix = to_sparse(adata.layers[layer] if layer is not None else adata.X)
    # write mtx file
    ul.log(__name__).info(f"Write mtx file")
    import scipy.io as scio
    scio.mmwrite(os.path.join(dir_path, 'matrix.mtx'), sparse_matrix.T, field=field)

    # Cell annotation
    ul.log(__name__).info(f"Write cell annotation")
    cell_info: DataFrame = adata.obs
    cell_info["barcodes"] = adata.obs.index.to_list()
    cell_info.to_csv(
        os.path.join(dir_path, "annotation.txt"),
        index=False,
        sep="\t",
        lineterminator="\n",
        encoding="utf-8"
    )

    # barcodes
    ul.log(__name__).info(f"Write barcodes")
    barcodes = pd.DataFrame(adata.obs.index.to_list(), columns=["index"])
    barcodes.to_csv(
        os.path.join(dir_path, "barcodes.tsv"),
        index=False,
        header=False,
        sep="\t",
        lineterminator="\n",
        encoding="utf-8"
    )

    # feature
    ul.log(__name__).info(f"Write feature")
    feature_info: DataFrame = adata.var
    if feature_name.split(".")[0] == "peaks":
        feature = pd.DataFrame(feature_info.index.to_list(), columns=["index"])
        new_feature = feature["index"].astype(str).str.split("[:-]", expand=True)
        new_feature.to_csv(
            os.path.join(dir_path, feature_name),
            index=False,
            header=False,
            sep="\t",
            lineterminator="\n",
            encoding="utf-8"
        )
    else:
        feature = pd.DataFrame(feature_info.index.to_list(), columns=["index"])
        feature.to_csv(
            os.path.join(dir_path, feature_name),
            index=False,
            header=False,
            sep="\t",
            lineterminator="\n",
            encoding="utf-8"
        )


def to_fragments(
    adata: AnnData,
    fragments: str,
    layer: str = None,
    is_sort: bool = True,
    is_gz: bool = True,
    is_keep: bool = False
) -> None:
    """
    Convert AnnData format data into fragments format file.
    :param adata: input data;
    :param fragments: output file name;
    :param layer: The layer of data that needs to form fragments file;
    :param is_sort: If set to true, the output will be sorting of chromatin;
    :param is_gz: Whether to compress the formed file to gz;
        Compress fragments file using method `pysam.tabix_compress`.
    :param is_keep: Set to true, The `fragments.tsv` file will not be retained after compression;
        When is_gz is true, it takes effect.
    :return: fragments format file;
    """

    output_path = os.path.dirname(fragments)
    ul.file_method(__name__).makedirs(output_path)

    data = check_adata_get(adata=adata, layer=layer, is_dense=False, is_matrix=False).T

    # get group information
    data_obs: DataFrame = data.obs.copy()
    data_var: DataFrame = data.var.copy()

    if "chr" not in data_obs.columns or "start" not in data_obs.columns or "end" not in data_obs.columns:
        ul.log(__name__).error("`chr` or `start`or  `end` not in obs column")
        raise ValueError("`chr` or `start` or `end` not in obs column")

    if "barcodes" not in data_var.columns:
        ul.log(__name__).error(f"`barcodes` not in obs column")
        raise ValueError(f"`barcodes` not in obs column")

    if is_sort:
        ul.log(__name__).info("Sort chromatin")
        data_obs["chr"] = data_obs["chr"].astype(chrtype)
        data_obs.sort_values(["chr", "start"], inplace=True)
        source_row_size = data_obs.shape[0]
        data_obs.dropna(subset=['chr'])

        if source_row_size > data_obs.shape[0]:
            chrs_str = ",".join(list(chrtype.categories))
            ul.log(__name__).warning(f"The chromatin with `chr` not in `{chrs_str}` has been deleted here.")

        data = data[data_obs.index, :]

    matrix = to_sparse(data.X, is_matrix=False)

    row_size, col_size = data.shape
    row_range, col_range = range(row_size), range(col_size)

    # Convert to dictionary
    barcodes_dict: dict = dict(zip(list(col_range), data_var.index))
    peaks_dict: dict = dict(zip(list(row_range), zip(data_obs["chr"], data_obs["start"], data_obs["end"])))

    nonzero = matrix.nonzero()
    nonzero_size = nonzero[0].size
    ul.log(__name__).info(f"Get size {row_size, col_size} ===> nonzero size: {nonzero_size}")

    ul.log(__name__).info(f"Generate the `fragments` file {fragments}.")
    with open(fragments, mode="w", encoding="utf-8", newline="\n") as f:

        f.write(f"# output_file = {fragments}\n")
        f.write(f"# layer = {layer}\n")
        f.write(f"# features: {row_size}, barcodes: {col_size}, nonzero: {nonzero_size}\n")

        for row, col in tqdm(zip(nonzero[0], nonzero[1])):

            # info
            peaks = peaks_dict[row]
            barcodes = barcodes_dict[col]
            f.write(f"{peaks[0]}\t{peaks[1]}\t{peaks[2]}\t{barcodes}\t{matrix[row, col]}\n")

    if is_gz:
        is_success: bool = True

        try:
            ul.log(__name__).info(f"Generate the `fragments` file {fragments}.gz.")
            import pysam
            pysam.tabix_compress(fragments, f"{fragments}.gz", force=True)
        except Exception as e:
            is_success: bool = False
            ul.log(__name__).error(f"Compression of {fragments} file failed. \n {e}")

        if is_success and not is_keep:
            os.remove(fragments)
